/*
 *  PyHID.cpp
 *
 *  Created by Canis Lupus
 *  Copyright 2006 Wooji Juice. All rights reserved.
 *
 */

#include <Python.h>
#include <vector>
#include <memory>
#include "HID.h"

//---------------------------------------------------------------------------
// Error-checking #defines

// Raise a Python exception if the HID Manager hasn't been initialised:
#define ENSURE_INIT \
    if (!TheHIDManager) \
        { PyErr_SetString(PyExc_RuntimeError,"PyHID must be initialised first."); return NULL; }
        
// Raise a Python exception if the device index passed in is invalid:
#define VALIDATE_DEVICE_INDEX(index) \
    if ((index<0)||(index>=TheHIDManager->GetDeviceCount())) \
        { PyErr_SetString(PyExc_ValueError,"Device index out of range."); return NULL; }

// Raise a Python exception if the element index passed in is invalid:
#define VALIDATE_ELEMENT_INDEX(device, index) \
    if ((index<0)||(index>=device->GetElementCount())) \
        { PyErr_SetString(PyExc_ValueError,"Element index out of range."); return NULL; }

// Catch any HID C++ exceptions that occur, convert them to Python exceptions and raise those:
#define CATCH_AND_CONVERT_EXCEPTIONS \
    catch (const HIDException& e) \
    { PyErr_SetString(PyExc_SystemError, e.GetReason().c_str()); return NULL; }


//---------------------------------------------------------------------------
// Globals

static HIDManager* TheHIDManager = NULL;


//---------------------------------------------------------------------------
// Init/fini

static PyObject *
PyHID_initalise(PyObject *self, PyObject *args)
{
    PyObject* result = Py_None;
    if (!TheHIDManager) 
    {
        try
        {
            TheHIDManager = new HIDManager();
        }
        CATCH_AND_CONVERT_EXCEPTIONS;
    }
    Py_INCREF(result);
    return result;
}


static PyObject *
PyHID_finalise(PyObject *self, PyObject *args)
{
    delete TheHIDManager;
    TheHIDManager = NULL;
    Py_INCREF(Py_None);
    return Py_None;
}


// Module-level

static PyObject *
PyHID_get_device_count(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    return Py_BuildValue("i", TheHIDManager->GetDeviceCount());
}

static PyObject *
PyHID_rescan_devices(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    try
    {
        TheHIDManager->ScanDevices();
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}


// Device-level

static PyObject *
PyHID_get_device_name(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    return Py_BuildValue("s", TheHIDManager->GetDevice(index)->GetDeviceName().c_str());
}

static PyObject *
PyHID_get_device_class_name(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    return Py_BuildValue("s", TheHIDManager->GetDevice(index)->GetDeviceClassName().c_str());
}

static PyObject *
PyHID_get_device_manufacturer(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    return Py_BuildValue("s", TheHIDManager->GetDevice(index)->GetManufacturerName().c_str());
}

static PyObject *
PyHID_get_device_element_count(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    return Py_BuildValue("i", TheHIDManager->GetDevice(index)->GetElementCount());
}

static PyObject *
PyHID_enable_monitoring(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, depth=8;
    if (!PyArg_ParseTuple(args, "I|I", &index, &depth))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    try
    {
        TheHIDManager->GetDevice(index)->StartQueue(depth);
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
PyHID_disable_monitoring(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    try
    {
        TheHIDManager->GetDevice(index)->StopQueue();
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
PyHID_poll_device(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index;
    if (!PyArg_ParseTuple(args, "I", &index))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    try
    {
        IHIDElement* element;
        int raw_value;
        TimeUnits time;
        bool hadEvent = TheHIDManager->GetDevice(index)->PollQueue(element, raw_value, time);
        if (hadEvent)
        {
            return Py_BuildValue("iif", element->GetIndex(), raw_value, time);
        }
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}


// Element level

static PyObject *
PyHID_get_element_usage(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    return Py_BuildValue("i", device->GetElement(element)->GetUsage());
}

static PyObject *
PyHID_get_element_usage_page(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    return Py_BuildValue("i", device->GetElement(element)->GetUsagePage());
}

static PyObject *
PyHID_get_element_value(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        return Py_BuildValue("i", device->GetElement(element)->GetValue());
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
}

static PyObject *
PyHID_set_element_value(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    int value;
    if (!PyArg_ParseTuple(args, "IIi", &index, &element, &value))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        device->GetElement(element)->SetValue(value);
        Py_INCREF(Py_None);
        return Py_None;
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
}

static PyObject *
PyHID_get_element_value_min(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        return Py_BuildValue("i", device->GetElement(element)->GetMinValue());
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
}

static PyObject *
PyHID_get_element_value_max(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        return Py_BuildValue("i", device->GetElement(element)->GetMaxValue());
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
}

static PyObject *
PyHID_get_element_value_normalised(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        return Py_BuildValue("f", device->GetElement(element)->GetNormalisedValue());
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
}

static PyObject *
PyHID_set_element_normalisation_range(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    float min, max;
    if (!PyArg_ParseTuple(args, "IIff", &index, &element, &min, &max))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        device->GetElement(element)->SetNormalisedRange(min, max);
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
PyHID_start_monitoring_element(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        device->StartMonitoringElement(element);
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
PyHID_stop_monitoring_element(PyObject *self, PyObject *args)
{
    ENSURE_INIT;
    unsigned int index, element;
    if (!PyArg_ParseTuple(args, "II", &index, &element))
        return NULL;
    VALIDATE_DEVICE_INDEX(index);
    IHIDevice* device = TheHIDManager->GetDevice(index);
    VALIDATE_ELEMENT_INDEX(device, element);
    try
    {
        device->StopMonitoringElement(element);
    }
    CATCH_AND_CONVERT_EXCEPTIONS;
    Py_INCREF(Py_None);
    return Py_None;
}


//---------------------------------------------------------------------------
// Python module goo

static PyMethodDef PyHIDMethods[] = {
    {"initialise",                  PyHID_initalise,                    METH_VARARGS,
		"Initialise the HID drivers."},
    {"finalise",                    PyHID_finalise,                     METH_VARARGS,
		"Shut down the HID drivers."},
		
	{"rescan_devices",              PyHID_rescan_devices,               METH_VARARGS,
	    "Rescan for connected HIDevices."},
	{"get_device_count",            PyHID_get_device_count,             METH_VARARGS,
	    "Returns the number of connected HIDevices."},
	
	{"get_device_name",             PyHID_get_device_name,              METH_VARARGS,
	    "Returns the name of the given device."},
	{"get_device_class_name",       PyHID_get_device_class_name,        METH_VARARGS,
	    "Returns the USB HID class name of the given device."},
	{"get_device_manufacturer",     PyHID_get_device_manufacturer,      METH_VARARGS,
	    "Returns the name of the manufacturer of the given device."},
	{"get_device_element_count",    PyHID_get_device_element_count,     METH_VARARGS,
	    "Returns the number of elements on the given device."},
	{"enable_monitoring",           PyHID_enable_monitoring,            METH_VARARGS,
	    "Enable event monitoring and queuing for the given device."},
	{"disable_monitoring",          PyHID_disable_monitoring,           METH_VARARGS,
	    "Disable event monitoring and queuing for the given device."},
	{"poll_device",                 PyHID_poll_device,                  METH_VARARGS,
	    "Poll the device's event queue; returns (element index, value, timestamp) or None."},

	{"get_element_usage",           PyHID_get_element_usage,            METH_VARARGS,
	    "Returns the USB HID Usage of the device element."},
	{"get_element_usage_page",      PyHID_get_element_usage_page,       METH_VARARGS,
	    "Returns the USB HID Usage Page of the device element."},
	{"get_element_value",           PyHID_get_element_value,            METH_VARARGS,
	    "Returns the raw value of the device element."},
	{"set_element_value",           PyHID_set_element_value,            METH_VARARGS,
	    "Sets the raw value of the device element."},
	{"get_element_value_min",       PyHID_get_element_value_min,        METH_VARARGS,
	    "Returns the minimum value of the device element."},
	{"get_element_value_max",       PyHID_get_element_value_max,        METH_VARARGS,
	    "Returns the maximum value of the device element."},
	{"get_element_value_normalised",PyHID_get_element_value_normalised, METH_VARARGS,
	    "Returns the normalised value of the device element."},
	{"set_element_normalisation_range",PyHID_set_element_normalisation_range, METH_VARARGS,
	    "Sets the min and max values returned by get_element_value_normalised()."},
	{"start_monitoring_element",    PyHID_start_monitoring_element,     METH_VARARGS,
	    "Adds the device element to the list of elements monitored by the device's event queue."},
	{"stop_monitoring_element",     PyHID_stop_monitoring_element,      METH_VARARGS,
	    "Removes the device element to the list of elements monitored by the device's event queue."},

    {NULL, NULL, 0, NULL}        /* Sentinel */
};


PyMODINIT_FUNC
initlowlevel(void)
{
    (void) Py_InitModule("lowlevel", PyHIDMethods);
}
